//**************************************************************************
//  Copyright (c) 2011 Igor Timurov, t051200@yandex.ru
//  All Rights Reserved.
//**************************************************************************

#ifndef __FMATRIX3__HED__
#define __FMATRIX3__HED__

#include "FBase.h"
#include "FVector3.h"

#pragma pack(push,1)
struct FMatrix3
{
    union 
    {
        struct 
        {
            float _11, _12, _13;
            float _21, _22, _23;
            float _31, _32, _33;
        };
        float m[3][3];
    };

    FMatrix3():
        _11(0), _12(0), _13(0),
        _21(0), _22(0), _23(0),
        _31(0), _32(0), _33(0)
    {}

    FMatrix3(
        float in_11, float in_12, float in_13,
        float in_21, float in_22, float in_23,
        float in_31, float in_32, float in_33):
        _11(in_11), _12(in_12), _13(in_13),
        _21(in_21), _22(in_22), _23(in_23),
        _31(in_31), _32(in_32), _33(in_33)
    {}

    FMatrix3(const FMatrix3& in):
        _11(in._11), _12(in._12), _13(in._13),
        _21(in._21), _22(in._22), _23(in._23),
        _31(in._31), _32(in._32), _33(in._33)
    {}

    void SetRow(const FVector3& v1, const FVector3& v2, const FVector3& v3)
    {
        _11 = v1.X, _12 = v1.Y, _13 = v1.Z;
        _21 = v2.X, _22 = v2.Y, _23 = v2.Z;
        _31 = v3.X, _32 = v3.Y, _33 = v3.Z;
    }

    void SetCol(int n, const FVector3& v)
    {
        m[0][n] = v.X;
        m[1][n] = v.Y;
        m[2][n] = v.Z;
    }

    void SetRow(int n, const FVector3& v)
    {
        m[n][0] = v.X;
        m[n][1] = v.Y;
        m[n][2] = v.Z;
    }

    inl void Transpose()
    {
        swap(_21, _12);
        swap(_31, _13);
        swap(_32, _23);
    }
    
    inl FMatrix3& operator = (const FMatrix3& in)
    {
        _11 = in._11, _12 = in._12, _13 = in._13;
        _21 = in._21, _22 = in._22, _23 = in._23;
        _31 = in._31, _32 = in._32, _33 = in._33;
        return *this;
    }

    inl FMatrix3& operator *= (const FMatrix3& in)
    {
        FMatrix3 a(*this);
        const FMatrix3* _in;
        if (this == &in)
            _in = &a;
        else
            _in = &in;
        const FMatrix3& b = *_in;

        #define __MULT(x,y)                                   \
            FDF_CONCAT3(_, y, x) =                            \
            FDF_CONCAT3(a._, y, 1)*FDF_CONCAT3(b._, 1, x) +   \
            FDF_CONCAT3(a._, y, 2)*FDF_CONCAT3(b._, 2, x) +   \
            FDF_CONCAT3(a._, y, 3)*FDF_CONCAT3(b._, 3, x);
        __MULT(1,1);__MULT(2,1);__MULT(3,1);
        __MULT(1,2);__MULT(2,2);__MULT(3,2);
        __MULT(1,3);__MULT(2,3);__MULT(3,3);
        #undef __MULT
        return *this;
    }

    inl FMatrix3& operator *= (float in)
    {
        _11 *= in, _12 *= in, _13 *= in;
        _21 *= in, _22 *= in, _23 *= in;
        _31 *= in, _32 *= in, _33 *= in;
        return *this;
    }

    inl FMatrix3& operator /= (float in)
    {
        *this *= 1.0f/in;
        return *this;
    }

    inl FMatrix3& Identity()
    {
        _11 = 1, _12 = 0, _13 = 0;
        _21 = 0, _22 = 1, _23 = 0;
        _31 = 0, _32 = 0, _33 = 1;
        return *this;
    }

    inl FMatrix3& Rotation(float x, float y, float z)
    {
        float cx = cos(z);
        float sx = sin(z);
        float cy = cos(y);
        float sy = sin(y);
        float cz = cos(x);
        float sz = sin(x);
        _11 = cx*cy;
        _12 = -sx*cz + cx*sy*sz;
        _13 = cx*sy*cz + sx*sz;
        _21 = sx*cy;
        _22 = sx*sy*sz + cx*cz;
        _23 = -cx*sz + sx*sy*cz;
        _31 = -sy;
        _32 = cy*sz;
        _33 = cy*cz;
        return *this;
    }

    inl float Determinant()
    {
        return _11*(_22*_33 - _23*_32) - _12*(_21*_33 - _23*_31) + _13*(_21*_32 - _22*_31);
    }

    inl void Invertible()
    {
        float d = Determinant();
        FMatrix3 t(*this);
        _11 = t._22*t._33 - t._23*t._32;
        _12 = t._23*t._31 - t._21*t._33;
        _13 = t._21*t._32 - t._22*t._31;

        _21 = t._13*t._32 - t._12*t._33;
        _22 = t._11*t._33 - t._13*t._31;
        _23 = t._12*t._31 - t._11*t._32;

        _31 = t._12*t._23 - t._13*t._22;
        _32 = t._13*t._21 - t._11*t._23;
        _33 = t._11*t._22 - t._12*t._21;
        Transpose();
        *this /= d;
    }

    inl void Transform(FVector3& a) const
    {
        float x = a.X*_11 + a.Y*_21 + a.Z*_31;
        float y = a.X*_12 + a.Y*_22 + a.Z*_32;
        float z = a.X*_13 + a.Y*_23 + a.Z*_33;
        a.X = x;
        a.Y = y;
        a.Z = z;
    }
};
#pragma pack(pop)

#endif //__FMATRIX3__HED__