//**************************************************************************
//  Copyright (c) 2011 Igor Timurov, t051200@yandex.ru
//  All Rights Reserved.
//**************************************************************************

#ifndef __FSCANNER__HED__
#define __FSCANNER__HED__

#include "FMatrix.h"
#include "FColor.h"
#include "FScannerImage.h"
#include "FScannerCeil.h"
#include "FAlgorithm.h"
#include "FMatrix3.h"

class FScanner
{
public:
    ULong width;
    ULong height;
    FMatrix<FScannerCeil> map;
    vector<FScannerImage*> images;

    float Ka; //    (  )
    float Kd; //   ( )
    float Ks; //   ( )
    float Ns; //    ( )
    float Bs; // Base size

public:
    FScanner(ULong _width, ULong _height)
    {
        width = _width;
        height = _height;
        Ka = 0.0f;
        Kd = 1.0f;
        Ks = 0.0f;
        Ns = 0.0f;
        Bs = 1.0f;
        map.Create(width, height);
    }

    ~FScanner()
    {
        for(ULong i = 0; i < images.size(); ++i)
            delete images[i];
        images.clear();
    }

    ULong AddImage(FScannerImage* info)
    {
        images.push_back(info);
        return images.size() - 1;
    }

    ULong Width()
    {
        return width;
    }

    ULong Height()
    {
        return height;
    }

    void MakeNormal(ULong img1, ULong img2, ULong img3)
    {
        FScannerImage& si1 = *images[img1];
        FScannerImage& si2 = *images[img2];
        FScannerImage& si3 = *images[img3];
        FScannerCeil** t = map.Src();

        FMatrix3 m;
        m.SetRow(si1.direct, si2.direct, si3.direct);
        FMatrix3 me(m);
        m.Invertible();
        m.Transpose();
        me.Transpose();

        for(ULong y = 0; y < height; ++y)
        {
            for(ULong x = 0; x < width; ++x)
            {
                // get data
                FScannerCeil& s = t[y][x];
                FVector3 c;
                c.X = si1.gray.Src()[y][x];
                c.Y = si2.gray.Src()[y][x];
                c.Z = si3.gray.Src()[y][x];

                FVector3& a = s.normal;
                a = c;
                m.Transform(a);
                a.Normalize();

                s.error = a;
                me.Transform(s.error);
                s.error -= c;
            }
        }
    }

    void MakeDelta()
    {
        int h = (int)map.Height();
        int w = (int)map.Width();
        FScannerCeil** t = map.Src();

        for(int y = 0; y < h; ++y)
        {
            for(int x = 0; x < w; ++x)
            {
                FScannerCeil& s = t[y][x];
                float z = 1.0f / s.normal.Z;
                //float z = sqrt(2.0f);
                if (z < 0.1f)
                    z = 0.1f;
                s.deltaX = -s.normal.X * z;
                s.deltaY = -s.normal.Y * z;
            }
        }
    }

    void MakeOneIteration()
    {
        MakeDeltaIteration(true);
        MakeDeltaIteration(false);
    }

    void MakeDeltaIteration(bool even)
    {
        int h = (int)map.Height();
        int w = (int)map.Width();
        FScannerCeil** t = map.Src();

        for(int y = 1; y < h; ++y)
        {
            int begin = even ? 2 : 1;
            even = !even;
            for(int x = begin; x < w; x += 2)
            {
                float& a = t[y-1][x-1].deltaX;
                float& b = t[y-1][x].deltaY;
                float& c = t[y-1][x-1].deltaY;
                float& d = t[y][x-1].deltaX;

                float u = a + b;
                float w = c + d;

                float iter = (u - w) / 4;
                a -= iter;
                b -= iter;
                c += iter;
                d += iter;
            }
        }
    }

    void MakeHeight()
    {
        int h = (int)map.Height();
        int w = (int)map.Width();
        FScannerCeil** t = map.Src();

        t[0][0].height = 0;
        for(int x = 1; x < w; ++x)
        {
            t[0][x].height = t[0][x-1].height + t[0][x-1].deltaX;
        }

        for(int y = 1; y < h; ++y)
        {
            t[y][0].height = t[y-1][0].height + t[y-1][0].deltaY;
            for(int x = 1; x < w; ++x)
            {
                float a = t[y][x-1].height + t[y][x-1].deltaX;
                float b = t[y-1][x].height + t[y-1][x].deltaY;
                t[y][x].height = (a + b) / 2;
            }
        }
    }

    void MakeHeight2()
	{
		int h = (int)map.Height();
        int w = (int)map.Width();
        FScannerCeil** t = map.Src();

        for(int y = 0; y < h; ++y)
        {
            for(int x = 1; x < w; ++x)
            {
                t[y][x].height = t[y][x-1].height + t[y][x-1].deltaX;
            }
        }
    }

    void GrayToBmp(ULong img, UByte* trg)
    {
        FScannerImage& si = *images[img];
        ULong bytesPerLine = BytesPerLine(width, 24);
        for(ULong y = 0; y < height; ++y)
        {
            for(ULong x = 0; x < width; ++x)
            {
                UByte* c = trg + y * bytesPerLine + x * 3;
                UByte g = (UByte)(si.gray.Src()[y][x] * 255.0f);
                GrayTo(c, g);
            }
        }
    }

    void NormalToBmp(UByte* trg)
    {
        ULong bytesPerLine = BytesPerLine(width, 24);
        FScannerCeil** t = map.Src();
        for(ULong y = 0; y < height; ++y)
        {
            for(ULong x = 0; x < width; ++x)
            {
                UByte* c = trg + y * bytesPerLine + x * 3;
                FVector3& a = t[y][x].normal;
                c[2] = (UByte)((a.X+1) * 128.0f);
                c[1] = (UByte)((a.Y+1) * 128.0f);
                c[0] = (UByte)((a.Z+1) * 128.0f);
            }
        }
    }

    void NormalToDxyBmp(UByte* trg)
    {
        ULong bytesPerLine = BytesPerLine(width, 24);
        FScannerCeil** t = map.Src();
        for(ULong y = 0; y < height; ++y)
        {
            for(ULong x = 0; x < width; ++x)
            {
                UByte* c = trg + y * bytesPerLine + x * 3;
                FVector3& a = t[y][x].normal;
                DirectToRGB(a.X, a.Y, c);
            }
        }
    }

    void NormalToDzBmp(UByte* trg)
    {
        ULong bytesPerLine = BytesPerLine(width, 24);
        FScannerCeil** t = map.Src();
        for(ULong y = 0; y < height; ++y)
        {
            for(ULong x = 0; x < width; ++x)
            {
                UByte* c = trg + y * bytesPerLine + x * 3;
                FVector3& a = t[y][x].normal;
                DirectHToRGB(a.Z, c);
            }
        }
    }
    
    void ErrorToBmp(UByte* trg)
    {
        ULong bytesPerLine = BytesPerLine(width, 24);
        FScannerCeil** t = map.Src();
        for(ULong y = 0; y < height; ++y)
        {
            for(ULong x = 0; x < width; ++x)
            {
                FVector3 v = map.Src()[y][x].error;

                UByte* c = trg + y * bytesPerLine + x * 3;
                c[0] = (UByte)((0.5f+v.X)*255.0f);
                c[1] = (UByte)((0.5f+v.Y)*255.0f);
                c[2] = (UByte)((0.5f+v.Z)*255.0f);
            }
        }
    }

    void HeightToBmp(UByte* trg)
    {
        ULong bytesPerLine = BytesPerLine(width, 24);
        FScannerCeil** t = map.Src();
        float maxv = 0; 
		float minv = 0;
		for(ULong j = 0; j < map.Height(); ++j)
		{
			for(ULong i = 0; i < map.Width(); ++i)
			{
				float v = map.Src()[j][i].height;
				maxv = max(maxv, v);
				minv = min(minv, v);
			}
		}
		float d = maxv - minv;
		if (d != 0)
		{
			for(ULong y = 0; y < height; ++y)
			{
				for(ULong x = 0; x < width; ++x)
				{
                    UByte* c = trg + y * bytesPerLine + x * 3;
					float v = map.Src()[y][x].height;
					v = ((v - minv)/d)*255.0f;
					GrayTo(c, (UByte)v);
				}
			}
		}

        //ofstream out("output.txt");
        //for(ULong x = 0; x < width; ++x)
        //{
        //    float v = map.Src()[0][x].height;
        //    v = ((v - minv)/d);
        //    out << v << "\n";
        //}
    }

    void HeightToRaw(UShort* trg, float& scale)
    {
        FScannerCeil** t = map.Src();
        float maxv = 0; 
		float minv = 0;
		for(ULong j = 0; j < map.Height(); ++j)
		{
			for(ULong i = 0; i < map.Width(); ++i)
			{
				float v = map.Src()[j][i].height;
				maxv = max(maxv, v);
				minv = min(minv, v);
			}
		}
		float d = maxv - minv;
		if (d != 0)
		{
            UShort* c = trg;
			for(ULong y = 0; y < height; ++y)
			{
				for(ULong x = width-1; x < width; --x)
				{
					float v = map.Src()[y][x].height;
					v = ((v - minv) / d) * 65535.0f;
					*c = (UShort)v;
                    ++c;
				}
			}
            scale = abs(d);
		}
    }

    void HeightToFloat(float* trg, float& scale)
    {
        FScannerCeil** t = map.Src();
        float maxv = 0; 
		float minv = 0;
		for(ULong j = 0; j < map.Height(); ++j)
		{
			for(ULong i = 0; i < map.Width(); ++i)
			{
				float v = map.Src()[j][i].height;
				maxv = max(maxv, v);
				minv = min(minv, v);
			}
		}
		float d = maxv - minv;
		if (d != 0)
		{
            float* c = trg;
			for(ULong y = 0; y < height; ++y)
			{
				for(ULong x = 0; x < width; ++x)
				{
					float v = map.Src()[y][x].height;
					*c = ((v - minv) / d);
                    ++c;
				}
			}
            scale = abs(d);
		}
    }

    /*

    void MakeDeltaIteration2(bool even)
    {
        int h = (int)map.Height();
        int w = (int)map.Width();
        FScannerCeil** t = map.Src();

        for(int y = 1; y < h; ++y)
        {
            int begin = even ? 2 : 1;
            even = !even;
            for(int x = begin; x < w; x += 2)
            {
                float& a = t[y-1][x-1].deltaX;
                float& b = t[y-1][x].deltaY;
                float& c = t[y-1][x-1].deltaY;
                float& d = t[y][x-1].deltaX;

                float u = a + b;
                float w = c + d;

                float iter = 0.01f;//abs((u - w) / 2);
                float iu;
                float iw;

                if (u < w)
                {
                    if (0 < w)
                    {
                        if (0 < u) // 0 < u < w
                        {
                            iu = 1.0f + iter;
                            iw = 1.0f - iter;
                        }
                        else // u < 0 < w
                        {
                            iu = 1.0f - iter;
                            iw = 1.0f - iter;
                        }
                    }
                    else // u < w < 0
                    {
                        iu = 1.0f - iter;
                        iw = 1.0f + iter;
                    }
                }
                else // w < u
                {
                    if (0 < u)
                    {
                        if (0 < w) // 0 < w < u
                        {
                            iu = 1.0f - iter;
                            iw = 1.0f + iter;
                        }
                        else // w < 0 < u
                        {
                            iu = 1.0f - iter;
                            iw = 1.0f - iter;
                        }
                    }
                    else // w < u < 0
                    {
                        iu = 1.0f + iter;
                        iw = 1.0f - iter;
                    }
                }

                a *= iu;
                b *= iu;
                c *= iw;
                d *= iw;
            }
        }
    }

    void MakeHeight4()
	{
		int h = (int)map.Height();
        int w = (int)map.Width();
        FScannerCeil** t = map.Src();

        for(int y = 0; y < h; ++y)
        {
            for(int x = 1; x < w; ++x)
            {
                t[y][x].height = t[y][x-1].height + t[y][x-1].norma.X;
            }
        }
        
        for(int x = 0; x < w; ++x)
        {
            for(int y = 1; y < h; ++y)
            {
                t[y][x].heightX += t[y-1][x].norma.Y;
            }
        }
	}

    void MakeHeight2()
	{
		int h = (int)map.Height();
        int w = (int)map.Width();
        FScannerCeil** t = map.Src();

        for(int y = 0; y < h; ++y)
        {
            for(int x = 1; x < w; ++x)
            {
                t[y][x].heightX = t[y][x-1].heightX + t[y][x-1].norma.X;
            }
        }
        
        for(int x = 0; x < w; ++x)
        {
            for(int y = 1; y < h; ++y)
            {
                t[y][x].heightY = t[y-1][x].heightY - t[y-1][x].norma.Y;
            }
        }

        for(int y = 0; y < h; ++y)
        {
            for(int x = 0; x < w; ++x)
            {
                t[y][x].height = (t[y][x].heightX + t[y][x].heightY)/2;
            }
        }
	}

    void MakeHeight3()
	{
		int h = (int)map.Height();
        int w = (int)map.Width();
        FScannerCeil** t = map.Src();

		int m = w+h;
		for(int k = 0; k < m; k++)
		{
			for(int x = 0; x <= k; x++)
			{
				int y = k - x;
				if (0 <= y && y < h && x < w)
				{
					t[y][x].height = GetHeight(x, y);
				}
			}
		}
	}

	float GetHeight(int x, int y)
	{
        FScannerCeil** t = map.Src();
		float h = 0;
		float d = 0;
		if(map.Check(x-1, y))
		{
			h += t[y][x-1].height + t[y][x].norma.X;
			d += 1.0;
		}
		if(map.Check(x, y-1))
		{
			h += t[y-1][x].height + t[y][x].norma.Y;
			d += 1.0;
		}
		return d == 0 ? 0 : h / d;
	}

    FScanner(const vector<FMapInfo>& info)
    {
        Ka = 0.0f;
        Kd = 1.0f;
        Ks = 0.0f;
        Ns = 0.0f;
        Bs = 1.0f;

        size = info.size();
        if (size == 0)
            throw 0;

        int width = info[0].image.Width();
        int height = info[0].image.Height();

        for(int i = 0; i < size; ++i)
        {
            if (width != info[i].image.Width() ||
                height != info[i].image.Height()) 
                throw 0;
        }

        map.Create(width, height);
        heightMap.Create(width, height);
        for(int y = 0; y < height; ++y)
        {
            for(int x = 0; x < width; ++x)
            {
                map.Src()[y][x].Init(size);
            }
        }

        for(int i = 0; i < size; ++i)
        {
            MakeGray(info[i].image, i);
            direct.push_back(info[i].direct);
        }

        MakeVGradient(2);
    }
    void Calculate()
    {
        MakeDelta();
        //MakeGradient();
        MakeHeight2();
    }
    void SaveGray(const wstring& path, int num)
    {
        FBitmap bmp;
        bmp.Make(map.Width(), map.Height(), 24);
        for(ULong j = 0; j < map.Height(); ++j)
        {
            for(ULong i = 0; i < map.Width(); ++i)
            {
				UByte val = map.Src()[j][i].gray[num];
				GrayTo(bmp.Pixel(i, j), val);
            }
        }
        bmp.Save(path);
    }

    void SaveDeltaX(const wstring& path)
    {
        FBitmap bmp;
        bmp.Make(map.Width(), map.Height(), 24);
        float maxv = 0; 
        float minv = 0;
        for(ULong j = 0; j < map.Height(); ++j)
        {
            for(ULong i = 0; i < map.Width(); ++i)
            {
				float v = map.Src()[j][i].deltaX;
                maxv = max(maxv, v);
                minv = min(minv, v);
            }
		}
		float d = maxv - minv;
		if (d != 0)
		{
			for(ULong j = 0; j < map.Height(); ++j)
			{
				for(ULong i = 0; i < map.Width(); ++i)
				{
					float v = map.Src()[j][i].deltaX;
					v = ((v - minv) / d);
					GrayTo(bmp.Pixel(i, j), (UByte)(v*255.0f));
				}
			}
		}
		bmp.Save(path);
	}
    
    void SaveDeltaY(const wstring& path)
    {
        FBitmap bmp;
        bmp.Make(map.Width(), map.Height(), 24);
        float maxv = 0; 
        float minv = 0;
        for(ULong j = 0; j < map.Height(); ++j)
        {
            for(ULong i = 0; i < map.Width(); ++i)
            {
				float v = map.Src()[j][i].deltaY;
                maxv = max(maxv, v);
                minv = min(minv, v);
            }
		}
		float d = maxv - minv;
		if (d != 0)
		{
			for(ULong j = 0; j < map.Height(); ++j)
			{
				for(ULong i = 0; i < map.Width(); ++i)
				{
					float v = map.Src()[j][i].deltaY;
					v = ((v - minv) / d);
					GrayTo(bmp.Pixel(i, j), (UByte)(v*255.0f));
				}
			}
		}
		bmp.Save(path);
	}

    void SaveHeiht(const wstring& path)
	{
		FBitmap bmp;
		bmp.Make(map.Width(), map.Height(), 24);
		float maxv = 0; 
		float minv = 0;
		for(ULong j = 0; j < map.Height(); ++j)
		{
			for(ULong i = 0; i < map.Width(); ++i)
			{
				float v = map.Src()[j][i].height;
				maxv = max(maxv, v);
				minv = min(minv, v);
			}
		}
		float d = maxv - minv;
		if (d != 0)
		{
			for(ULong j = 0; j < map.Height(); ++j)
			{
				for(ULong i = 0; i < map.Width(); ++i)
				{
					float v = map.Src()[j][i].height;
					v = ((v - minv) / d) * 255.0f;
					//GToRGB(bmp.Pixel(i, j), (UByte)v);
					GrayTo(bmp.Pixel(i, j), (UByte)v);
                    heightMap.Src()[j][i] = v;
				}
			}
		}
		bmp.Save(path);
    }

    void SaveRaw(const wstring& path)
    {
        FFileWriter file;
        file.Open(path);
        for(ULong j = 0; j < map.Height(); ++j)
        {
            for(ULong i = 0; i < map.Width(); ++i)
            {
                float v = heightMap.Src()[j][i];//65535.0f;
                file.WriteData((UByte)v);
            }
        }
        file.Close();
    }

    float GetAngle(float cNL1, float VL1, float cNL2, float VL2)
    {
        //return cNL2 - cNL1;
        // cNL1=[0,1]; VL1=[-PI/2,PI/2]; cNL2=[0,1]; VL2=[-PI/2,PI/2]
        float NL1 = acos(cNL1);
        float NL2 = acos(cNL2);
        float VN_L1a = VL1 - NL1; //   
        float VN_L1b = VL1 + NL1; //  
        float VN_L2b = VL2 - NL2; //  
        float VN_L2a = VL2 + NL2; //   

        float a = abs(VN_L1a - VN_L2a);
        float b = abs(VN_L1b - VN_L2b);
        float c = abs(VN_L1a - VN_L2b);
        float d = abs(VN_L1b - VN_L2a);
        
        float VN = 0;

        if(a < b && a < c && a < d) //  
        {
        }
        else if(b < c && b < d) // 
        {
            VN = (VN_L1b + VN_L2b) / 2;
        }
        else if(c < d) // 
        {
            VN = (VN_L1a + VN_L2b) / 2;
        }
        else // 
        {
            VN = (VN_L1b + VN_L2a) / 2;
        }

        float H = Bs * tan(VN);
        return H;
    }

    void MakeDelta()
    {
        int h = (int)map.Height();
        int w = (int)map.Width();
        //float pbase = 1.0;
        //float acl = FDF_PI/6;
        FScannerCeil** t = map.Src();
        for(int y = 0; y < h; ++y)
        {
            for(int x = 0; x < w; ++x)
            {
        //        for(int i = 0; i < size; ++i)
        //        {
        //            float v = (float)t[y][x].gray[i] / 255.0f;
				    //t[y][x].deltaX += v * direct[i].X;
				    //t[y][x].deltaY += v * direct[i].Y;
        //        }


                float v0 = (float)t[y][x].gray[0] / 255.0f;
                float v1 = (float)t[y][x].gray[1] / 255.0f;
                float v2 = (float)t[y][x].gray[2] / 255.0f;
                float v3 = (float)t[y][x].gray[3] / 255.0f;
                t[y][x].deltaX = GetAngle(v2, direct[2].X, v3, direct[3].X);
                t[y][x].deltaY = GetAngle(v0, direct[0].Y, v1, direct[1].Y);

            }
        }
    }
    */
};

#endif //__FSCANNER__HED__