﻿//**************************************************************************
//  Copyright (c) 2011 Igor Timurov, t051200@yandex.ru
//  All Rights Reserved.
//**************************************************************************

using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.Xml;
using System.Xml.Serialization;
using System.IO;
using System.Globalization;

namespace SurfaceScanner
{
    public class Scanner
    {
        delegate void ViewBitmap(BitmapData bmpData);

        IntPtr scannerContext = IntPtr.Zero;
        Rectangle imageRect = new Rectangle();
        List<ScannerImage> images = new List<ScannerImage>();

        public List<ScannerImage> Images
        {
            get { return images; }
        }

        public bool HasContext
        {
            get { return scannerContext != IntPtr.Zero; }
        }

        public Scanner()
        {
        }

        List<int> GetNumbers(string str)
        {
            List<int> v = new List<int>();
            foreach (string s in str.Split(' '))
            {
                int n = 0;
                if (int.TryParse(s, out n))
                    v.Add(n);
            }
            return v;
        }

        public void SetRect(string str)
        {
            List<int> v = GetNumbers(str);
            if (v.Count == 4)
            {
                imageRect.X = v[0];
                imageRect.Y = v[1];
                imageRect.Width = v[2];
                imageRect.Height = v[3];
            }
        }

        public string GetRect()
        {
            return string.Format("{0} {1} {2} {3}", imageRect.X, imageRect.Y, imageRect.Width, imageRect.Height);
        }

        Bitmap MakeBitmap(ViewBitmap view)
        {
            try
            {
                Bitmap bmp = new Bitmap(imageRect.Width, imageRect.Height, PixelFormat.Format24bppRgb);
                BitmapData bmpData = bmp.LockBits(new Rectangle(Point.Empty, imageRect.Size), ImageLockMode.WriteOnly, PixelFormat.Format24bppRgb);
                view(bmpData);
                bmp.UnlockBits(bmpData);
                return bmp;
            }
            catch
            {
                return null;
            }
        }

        public void Create()
        {
            Release();
            scannerContext = ScannerLib.CreateSC((uint)imageRect.Width, (uint)imageRect.Height);
            foreach(ScannerImage img in images)
                img.ToScanner(scannerContext, imageRect);
        }

        public void Dispose()
        {
            Release();
            foreach (ScannerImage si in images)
                si.Unload();
            images.Clear();
        }

        public void Release()
        {
            ScannerLib.ReleaseSC(scannerContext);
        }

        public ScannerImage LoadImage(string path)
        {
            ScannerImage image = new ScannerImage();
            image.Load(path);
            images.Add(image);
            imageRect.Width = image.Bmp.Width;
            imageRect.Height = image.Bmp.Height;
            return image;
        }

        public void MakeNormal(string str)
        {
            List<int> v = GetNumbers(str);
            UInt32 a1 = images[v[0]].Id;
            UInt32 a2 = images[v[1]].Id;
            UInt32 a3 = images[v[2]].Id;
            if (a1 != a2 && a2 != a3 && a1 != a3)
                ScannerLib.MakeNormal(scannerContext, a1, a2, a3);
        }

        public void CorrectNormal(ref int curv, int maxv)
        {
            ScannerLib.CorrectNormal(scannerContext, ref curv, maxv);
        }

        public void MakeHeight()
        {
            ScannerLib.MakeHeight(scannerContext);
        }

        public Bitmap GetNormal()
        {
            return MakeBitmap(delegate(BitmapData bmpData) { ScannerLib.NormalToBmp(scannerContext, bmpData.Scan0); });
        }

        public Bitmap GetNormalDxy()
        {
            return MakeBitmap(delegate(BitmapData bmpData) { ScannerLib.NormalToDxyBmp(scannerContext, bmpData.Scan0); });
        }

        public Bitmap GetNormalDz()
        {
            return MakeBitmap(delegate(BitmapData bmpData) { ScannerLib.NormalToDzBmp(scannerContext, bmpData.Scan0); });
        }

        public Bitmap GetError()
        {
            return MakeBitmap(delegate(BitmapData bmpData) { ScannerLib.ErrorToBmp(scannerContext, bmpData.Scan0); });
        }

        public Bitmap GetGray(UInt32 id)
        {
            return MakeBitmap(delegate(BitmapData bmpData) { ScannerLib.GrayToBmp(scannerContext, id, bmpData.Scan0); });
        }

        public Bitmap GetHeight()
        {
            return MakeBitmap(delegate(BitmapData bmpData) { ScannerLib.HeightToBmp(scannerContext, bmpData.Scan0); });
        }

        public void GetHeightData(out float[] matrix, out int width, out int height, out float scale)
        {
            scale = 0;
            width = imageRect.Width;
            height = imageRect.Height;
            matrix = new float[width * height];
            ScannerLib.HeightToFloat(scannerContext, matrix, ref scale);
        }

        public void SaveRaw(string path)
        {
            byte[] data = new byte[imageRect.Width * imageRect.Height * 2];
            float scale = 0;
            ScannerLib.HeightToRaw(scannerContext, data, ref scale);
            File.WriteAllBytes(path, data);
            File.WriteAllText(path + ".txt",
                string.Format("Heigth = {0}\nWidthX = {1}\nWidthY = {2}\nHeigth / WidthX = {3}\nHeigth / WidthY = {4}",
                scale,
                imageRect.Width,
                imageRect.Height,
                scale / imageRect.Width,
                scale / imageRect.Height));
        }

        static public Bitmap Sub(string path1, string path2)
        {
            Bitmap bmp1 = new Bitmap(path1);
            Bitmap bmp2 = new Bitmap(path2);
            BitmapData bmpData1 = bmp1.LockBits(new Rectangle(Point.Empty, bmp1.Size), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            BitmapData bmpData2 = bmp2.LockBits(new Rectangle(Point.Empty, bmp2.Size), ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb);
            ScannerLib.Sub(bmpData1.Scan0, bmpData2.Scan0, (UInt32)bmp1.Width, (UInt32)bmp1.Height);
            bmp1.UnlockBits(bmpData1);
            bmp2.UnlockBits(bmpData2);
            return bmp1;
        }

        #region Load/Save

        [Serializable]
        public class XmlData
        {
            [XmlIgnore]
            public Rectangle Rect = new Rectangle();

            public List<XmlImage> Images = new List<XmlImage>();
            public int X { get { return Rect.X; } set { Rect.X = value; } }
            public int Y { get { return Rect.Y; } set { Rect.Y = value; } }
            public int Width { get { return Rect.Width; } set { Rect.Width = value; } }
            public int Height { get { return Rect.Height; } set { Rect.Height = value; } }

            public void SetImages(List<ScannerImage> images)
            {
                foreach (ScannerImage i in images)
                {
                    XmlImage d = new XmlImage();
                    d.SetDirect(i.direct);
                    d.Path = i.Path;
                    this.Images.Add(d);
                }
            }
        }

        public class XmlImage 
        {
            public string Path;
            public float X;
            public float Y;
            public float Z;

            public void SetDirect(ScannerLib.Vector3 direct)
            {
                X = direct.X;
                Y = direct.Y; 
                Z = direct.Z;
            }

            public ScannerLib.Vector3 GetDirect()
            {
                ScannerLib.Vector3 direct = new ScannerLib.Vector3();
                direct.X = X;
                direct.Y = Y;
                direct.Z = Z;
                return direct;
            }
        }

        public void Save(string path)
        {
            Type t = typeof(List<XmlData>);
            XmlSerializer serializer = new XmlSerializer(typeof(XmlData));
            XmlData data = new XmlData();
            data.SetImages(images);
            data.Rect = imageRect;
            XmlWriter writer = new XmlTextWriter(path, System.Text.Encoding.UTF8);
            serializer.Serialize(writer, data);
            writer.Close();
        }

        static public Scanner Load(string path)
        {
            XmlSerializer serializer = new XmlSerializer(typeof(XmlData));
            XmlReader reader = new XmlTextReader(path);
            XmlData data = (XmlData)serializer.Deserialize(reader);
            reader.Close();
            Scanner scanner = new Scanner();
            foreach (XmlImage d in data.Images)
                scanner.LoadImage(d.Path).direct = d.GetDirect();
            scanner.imageRect = data.Rect;
            return scanner;
        }

        #endregion
    }
}
